# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Test utilies for Omicron
"""

import os

from .. import utils


def test_get_omicron_version():
    testv = 'v2r1'
    v = utils.get_omicron_version(
        "/home/detchar/opt/virgosoft/Omicron/%s/Linux-x86_64/omicron.exe"
        % testv
    )
    assert v == testv

    os.environ['OMICRON_VERSION'] = testv
    assert utils.get_omicron_version() == testv

    os.environ.pop('OMICRON_VERSION')
    os.environ['OMICRONROOT'] = (
        "/home/detchar/opt/virgosoft/Omicron/%s" % testv)
    assert utils.get_omicron_version() == testv
    assert utils.get_omicron_version() > 'v1r2'
    assert utils.get_omicron_version() < 'v2r2'



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Input/Output utilities for Omicron ROOT and LIGO_LW XML files
"""

import warnings
import os.path
import glob
import re
from collections import defaultdict

import numpy

from gwpy.io.cache import file_segment
from gwpy.time import tconvert

from . import const
from .segments import (Segment, segmentlist_from_tree)

re_delim = re.compile('[:_-]')


def merge_root_files(inputfiles, outputfile,
                     trees=['segments', 'triggers', 'metadata'],
                     strict=True, on_missing='raise'):
    """Merge several ROOT files into a single file

    Parameters
    ----------
    inputfile : `list` of `str`
        the paths of the input ROOT files to merge
    outputfile : `str`
        the path of the output ROOT file to write
    tree : `list` of `str`
        the names of the ROOT Trees to include
    strict : `bool`, default: `True`
        only combine contiguous files (as described by the contained segmenets)
    on_missing : `str`, optional
        what to do when an input file is not found, one of

        - ``'ignore'``: do nothing
        - ``'warn'``: print a warning
        - ``'raise'``: raise an `IOError`

    Notes
    -----
    This method requires the `ROOT <https://root.cern.ch/pyroot>`_ package.
    """
    import ROOT
    chains = {}

    # validate input files
    for f in inputfiles:
        missing = not os.path.isfile(f)
        msg = "No such file or directory: %r" % f
        if on_missing == 'ignore':
            pass
        elif missing and on_missing == 'warn':
            warnings.warn(msg)
        elif missing:
            raise IOError(msg)

    for tree in trees:
        chains[tree] = ROOT.TChain(tree)
        for i, f in enumerate(inputfiles):
            chains[tree].Add(f)
        if (strict and tree == 'segments' and
                len(segmentlist_from_tree(chains[tree]).coalesce()) > 1):
            raise RuntimeError("Cannot perform a 'strict' merge on files "
                               "containing discontiguous data")

    # write new file (use 'recreate' to overwrite old file)
    out = ROOT.TFile(outputfile, 'recreate')
    for i, (name, chain) in enumerate(chains.iteritems()):
        if i:
            out = ROOT.TFile(outputfile, 'update')  # reopen file
        chain.Merge(out, 0)
    return outputfile


def _parse_channel_and_filetag(channel, filetag):
    """Work out the relevant observatory and description given the inputs
    """
    cname = re_delim.sub('_', str(channel))
    obs, description = cname.split('_', 1)
    if filetag is not None:
        description += '_%s' % re_delim.sub('_', filetag).strip('_')
    return obs, description


def _iter_files_in_gps_directory(channel, basepath, gps5, ext,
                                 filetag=const.OMICRON_FILETAG.upper()):
    """Internal method to glob Omicron files from a directory structure
    """
    ifo, description = _parse_channel_and_filetag(channel, filetag)
    dg = os.path.join(basepath, ifo, description, str(gps5))
    for d in glob.iglob(dg):
        g = os.path.join(
            d, '%s-%s-*.%s' % (ifo, description, ext))
        for path in glob.iglob(g):
            yield path


def find_omicron_files(channel, start, end, basepath, ext='xml.gz',
                       filetag=const.OMICRON_FILETAG.upper()):
    """Find Omicron files under a given starting directory
    """
    gps5 = int(str(start)[:5])-1
    cache = list()
    span = Segment(start, end)
    while gps5 <= int(str(end)[:5]):
        new = _iter_files_in_gps_directory(channel, basepath, gps5,
                                           ext, filetag=filetag)
        cache.extend(path for path in new if
                     file_segment(path).intersects(span))
        gps5 += 1
    return cache


def find_latest_omicron_file(channel, basepath, ext='xml.gz',
                             filetag=const.OMICRON_FILETAG.upper(),
                             gps=None):
    """Find the most recent Omicron file for a given channel
    """
    if gps is None:
        gps = int(tconvert('now'))
    gps5 = int(str(gps)[:5])
    while gps5:
        cache = _iter_files_in_gps_directory(channel, basepath, gps5,
                                             ext, filetag=filetag)
        try:
            return list(cache)[-1]
        except IndexError:
            pass
        gps5 -= 1
    raise RuntimeError("Failed to find any Omicron files for %r" % channel)


def find_pending_files(channel, proddir, ext='xml.gz'):
    """Find files that have just been created, pending archival
    """
    ifo = channel.split(':', 1)[0]
    return glob.glob(os.path.join(
        proddir, 'triggers', channel, '%s-*.%s' % (ifo, ext)))


def get_archive_filename(channel, start, duration, ext='xml.gz',
                         filetag=const.OMICRON_FILETAG.upper(),
                         archive=const.OMICRON_ARCHIVE):
    """Returns the full file path for this channel's triggers

    This method will design a trigger file path for you, rather than find
    a file that is already there, and so should be used to seed an archive,
    not search it.

    Parameters
    ----------
    channel : `str`
        name of channel
    start : `int`
        GPS start time of file
    duration : `int`
        duration (seconds) of file
    ext : `str`, optional
        file extension, defaults to ``xml.gz``
    filetag : `str`, optional
        filetag to be appended after the channel name, defaults to ``OMICRON``
    archive : `str`, optional
        base directory of the trigger archive, defaults to
        `const.OMICRON_ARCHIVE`

    Returns
    -------
    filepath : `str`
        the absolute path where this file should be stored

    Notes
    -----
    See `T050017 <https://dcc.ligo.org/LIGO-T050017>`_ for details of the
    file-naming convention.

    Examples
    --------
    >>> get_archive_filename('H1:GDS-CALIB_STRAIN', 1234567890, 100, archive='/triggers')
    '/triggers/H1/GDS_CALIB_STRAIN_OMICRON/12345/H1-GDS_CALIB_STRAIN_OMICRON-1234567890-100.xml.gz'

    """  # noqa: E501
    ifo, description = _parse_channel_and_filetag(channel, filetag)
    filename = '%s-%s-%d-%d.%s' % (
        ifo, description, int(start), int(duration), ext)
    if start < 10000:
        gps5 = '%.5d' % int(start)
    else:
        gps5 = str(int(start))[:5]
    return os.path.join(archive, ifo, description, gps5, filename)


def merge_hdf5_files(inputfiles, outputfile, **compression_kw):
    """Merge several HDF5 files into a single file

    Parameters
    ----------
    inputfile : `list` of `str`
        the paths of the input HDF5 files to merge
    outputfile : `str`
        the path of the output HDF5 file to write
    tree : `list` of `str`
        the names of the HDF5 Trees to include
    strict : `bool`, default: `True`
        only combine contiguous files (as described by the contained segmenets)
    on_missing : `str`, optional
        what to do when an input file is not found, one of

        - ``'ignore'``: do nothing
        - ``'warn'``: print a warning
        - ``'raise'``: raise an `IOError`

    Notes
    -----
    This method requires the `HDF5 <https://root.cern.ch/pyroot>`_ package.
    """
    import h5py

    # get list of datasets
    attributes = {}
    datasets = {}
    for path in inputfiles:
        with h5py.File(path, 'r') as h5f:
            attributes = dict(h5f.attrs)
            for dset in h5f:
                # assert datatype is the same, and append shape
                shape = h5f[dset].shape
                dtype = h5f[dset].dtype
                try:
                    shape = numpy.sum(datasets[dset][0] + shape, keepdims=True)
                except KeyError:
                    chunk = shape
                else:
                    assert dtype == datasets[dset][1], (
                        "Cannot merge {0}/{1}, does not match dtype".format(
                            path, dset))
                    if chunk != datasets[dset][2]:
                        chunk = True
                datasets[dset] = (shape, dtype, chunk)

                # use default compression options from this file
                for copt in ('compression', 'compression_opts'):
                    compression_kw.setdefault(copt, getattr(h5f[dset], copt))

    # combine sets
    position = defaultdict(int)
    with h5py.File(outputfile, 'w') as h5out:
        # copy attributes (just from last file)
        for key in attributes:
            h5out.attrs[key] = attributes[key]
        # create datasets
        for dset, (shape, dtype, chunk) in datasets.items():
            h5out.create_dataset(dset, shape=shape, dtype=dtype,
                                 chunks=chunk, **compression_kw)
        # copy dataset contents
        for path in inputfiles:
            with h5py.File(path, 'r') as h5in:
                for dset in datasets:
                    data = h5in[dset]
                    size = data.shape[0]
                    pos = position[dset]
                    h5out[dset][pos:pos+size] = data
                    position[dset] += size

    return outputfile



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write/modify Omicron-format parameters files
"""

from __future__ import (division, print_function)

import os.path
import re
from math import (ceil, exp, floor, log)
from getpass import getuser
from datetime import datetime
try:  # python >= 3
    import configparser
except ImportError:  # python 2.x
    import ConfigParser as configparser

from ligo.segments import (segmentlist as SegmentList, segment as Segment)

from collections import OrderedDict

from . import (const, utils)
from .segments import integer_segments

CHANNEL_DELIM_REGEX = re.compile('[:_-]')
UNUSED_PARAMS = ['state-flag', 'frametype', 'state-channel', 'state-frametype']
OMICRON_PARAM_MAP = {
    'sample-frequency': ('DATA', 'SAMPLEFREQUENCY')
}


class OmicronParameters(configparser.ConfigParser):
    """Custom `configparser.ConfigParser` for Omicron parameters files
    """
    OMICRON_DEFAULTS = OrderedDict()
    OMICRON_DEFAULTS[None] = {
        'PARAMETER': {
            'CLUSTERING': 'TIME',
            'TRIGGERRATEMAX': 100000,
        },
        'OUTPUT': {
            'DIRECTORY': os.path.curdir,
            'PRODUCTS': 'triggers',
            'VERBOSITY': 1,
            'FORMAT': 'rootxml',
            'NTRIGGERMAX': 1e7,
        },
        'DATA': {
        },
    }
    OMICRON_DEFAULTS['v2r2'] = {
        'PARAMETER': {'FFTPLAN': 'FFTW_ESTIMATE'},
    }
    OMICRON_DEFAULTS['v2r3'] = {
        'OUTPUT': {'FORMAT': 'root xml hdf5'},
    }

    def __init__(self, version=None, defaults=dict(), **kwargs):
        configparser.ConfigParser.__init__(self, defaults=defaults, **kwargs)
        if version is None:
            try:
                version = utils.get_omicron_version()
            except KeyError:
                version = utils.OmicronVersion(const.OMICRON_VERSION)
        self.version = version
        self._set_defaults()

    def _set_defaults(self):
        """Set basic defaults for each config section
        """
        for version in self.OMICRON_DEFAULTS:
            if version is not None and self.version < version:
                continue
            for section, params in self.OMICRON_DEFAULTS[version].items():
                try:
                    self.add_section(section)
                except configparser.DuplicateSectionError:
                    pass
                for key, val in params.items():
                    if isinstance(val, tuple):
                        self.set(section, key, ' '.join(map(str, val)))
                    else:
                        self.set(section, key, str(val))

    # -- better option accessors ----------------

    def getlist(self, section, option):
        raw = self.get(section, option)
        return raw.split()

    def getfloats(self, section, option):
        return map(float, self.getlist(section, option))

    def optionxform(self, optionstr):
        return optionstr.upper()

    # -- input/output ---------------------------

    def _read(self, fp, fpname):
        """Read a file either either using INI or Omicron formatting
        """
        if fpname.endswith('.ini'):
            return configparser.ConfigParser._read(self, fp, fpname)
        channels = []
        for line in fp:
            if isinstance(line, bytes):
                line = line.decode()
            if not line.strip() or line[0] in '#;':  # blank
                continue
            sec, key, val = line.rstrip().split(None, 2)
            if sec.lower() == 'data' and key.lower() == 'channels':
                channels.append(val)
            else:
                self.set(sec, key, val)
        if channels:
            self.set('DATA', 'CHANNELS', ' '.join(channels))

    def write(self, fp):
        # write basic
        if fp.name.endswith('.ini'):
            return configparser.ConfigParser.write(self, fp)

        # print header
        print('# Omicron %s parameter file' % self.version, file=fp)
        print('# Written using pyomicron by %s at %s'
              % (getuser(), datetime.now()), file=fp)

        # write omicron-format
        def _write_option(sec, key, val):
            print('{0: <10}'.format(sec.upper()),
                  '{0: <16}'.format(key.upper()),
                  val, file=fp, sep=' ')

        for sec in self.sections():
            print("", file=fp)
            for key, val in self.items(sec):
                if sec.lower() == 'data' and key.lower() == 'channels':
                    for chan in val.split():
                        _write_option(sec, key, chan)
                else:
                    _write_option(sec, key, val)
    write.__doc__ = configparser.ConfigParser.write.__doc__

    def write_distributed(self, directory, nchannels=10):
        """Write multiple parameters files to distribute processing

        This method separates the channels list into blocks of ``nchannels``
        and writes a parameters file for each one into ``directory``.

        Parameters
        ----------
        directory : `str`
            the output directory for the parameters files

        nchannels : `int`, optional, default: ``10``
            the number of channels to insert in each parameters file

        Returns
        -------
        paramfile : `str`
            the file path of the complete parameters.txt file, written for
            reference
        parametermap : `dict`
            a `dict` of ``(filename, channel_list)`` pairs that indicate
            which written file contains which channels
        """
        # get channels list
        channels = self.getlist('DATA', 'CHANNELS')
        with open(os.path.join(directory, 'parameters.txt'), 'w') as f:
            self.write(f)
        tmpcp = self.__class__(version=self.version)
        tmpcp.read([f.name])

        # write files
        files = OrderedDict()
        i = 0
        while i * nchannels < len(channels):
            chans = channels[i * nchannels:(i+1) * nchannels]
            pfile = os.path.join(directory, 'parameters-%d.txt' % i)
            tmpcp.set('DATA', 'CHANNELS', ' '.join(chans))
            with open(pfile, 'w') as f2:
                tmpcp.write(f2)
            files[pfile] = chans
            i += 1
        return f.name, files

    @classmethod
    def from_channel_list_config(cls, config, section, version=None):
        """Read `OmicronParameters` from a LIGO channel list file configuration

        Parameters
        ----------
        config : `configparser.ConfigParser`, `str`
            either a configuration object, or a path to a config file

        section : `str`
            the name of the section to parse

        version : `str`, optional
           the version of Omicron for which this configuration is valid

        Returns
        -------
        parameters : `OmicronParameters`
            an Omicron-format version of the parameters from the configuration
        """
        new = cls(version=version)

        # open and parse config filename
        if isinstance(config, str):
            fn = config
            config = configparser.ConfigParser()
            config.read(fn)

        # reformat from separate low, high to a tuple of values
        for (range_, low_, high_) in [('frequency-range', 'flow', 'fhigh'),
                                      ('q-range', 'qlow', 'qhigh')]:
            if (not config.has_option(section, range_) and
                    config.has_option(section, low_)):
                config.set(
                    section, range_,
                    '%s %s' % (config.getfloat(section, low_),
                               config.getfloat(section, high_)))
            for opt in (low_, high_):
                try:
                    config.remove_option(section, opt)
                except configparser.NoOptionError:
                    pass

        for key, val in config.items(section):
            if key in UNUSED_PARAMS:
                continue
            elif key == 'channels':
                new.set('DATA', 'CHANNELS', ' '.join(val.split()))
            else:
                omikey = ''.join(key.split('-')).upper()
                try:
                    group, attr = OMICRON_PARAM_MAP[key]
                except KeyError:
                    if val in [None, 'None', 'none', '']:  # unset default
                        new.remove_option('PARAMETER', omikey)
                    else:
                        new.set('PARAMETER', omikey, val)
                else:
                    new.set(group, omikey, val)

        # get parameters
        if new.version >= 'v2r2':
            # rename 'CHUNKDURATION' -> 'PSDLENGTH'
            try:
                psdlength = new.get('PARAMETER', 'CHUNKDURATION')
            except configparser.NoOptionError:
                pass
            else:
                new.remove_option('PARAMETER', 'CHUNKDURATION')
                new.set('PARAMETER', 'PSDLENGTH', psdlength)
            # combine 'SEGMENTDURATION' and 'OVERLAPDURATION' into 'TIMING'
            try:
                timing = (new.get('PARAMETER', 'SEGMENTDURATION'),
                          new.get('PARAMETER', 'OVERLAPDURATION'))
            except configparser.NoOptionError:
                pass
            else:
                new.remove_option('PARAMETER', 'SEGMENTDURATION')
                new.remove_option('PARAMETER', 'OVERLAPDURATION')
                new.set('PARAMETER', 'TIMING', '%s %s' % timing)

        return new

    # -- utilities ------------------------------

    def validate(self):
        """Validate that Omicron will accept the segment parameters

        Parameters
        ----------
        chunk : `int`
            omicron CHUNKDURATION parameter
        segment : `int`
            omicron SEGMENTDURATION parameter
        overlap : `int`
            omicron OVERLAPDURATION parameter
        frange : `tuple` of `float`
            omicron FREQUENCYRANGE parameter

        Raises
        ------
        AssertionError
            if any of the parameters isn't acceptable for Omicron to process
        """
        # extract parameters (for convenience)
        try:
            sampling = self.getfloat('DATA', 'SAMPLEFREQUENCY')
        except configparser.NoOptionError:
            sampling = None
        try:
            frange = self.getfloats('PARAMETER', 'FREQUENCYRANGE')
        except configparser.NoOptionError:
            frange = None
        if self.version >= 'v2r2':
            chunk = self.getfloat('PARAMETER', 'PSDLENGTH')
            segment, overlap = self.getfloats('PARAMETER', 'TIMING')
        else:
            chunk = self.getfloat('PARAMETER', 'CHUNKDURATION')
            segment = self.getfloat('PARAMETER', 'SEGMENTDURATION')
            overlap = self.getfloat('PARAMETER', 'OVERLAPDURATION')

        # check parameters are valid
        assert segment <= chunk, (
            "Segment length is greater than chunk length")
        assert overlap <= segment, (
            "Overlap length is greater than segment length")
        assert overlap != 0, "Omicron cannot run with zero overlap"
        assert overlap % 2 == 0, "Padding (overlap/2) is non-integer"
        assert segment >= (2 * overlap), (
            "Overlap is too large, cannot be more than 50%")
        dchunk = chunk - overlap
        dseg = segment - overlap
        assert dchunk % dseg == 0, (
            "Chunk duration doesn't allow an integer number of segments, "
            "%ds too large" % (dchunk % dseg))
        if sampling is None or frange is None:
            return
        if frange[0] < 1:
            x = 10 * floor(sampling / frange[0])
            psdsize = 2 * int(2 ** ceil(log(x) / log(2.)))
        else:
            psdsize = 2 * sampling
        psdlen = psdsize / sampling
        chunkp = chunk * sampling
        overlapp = overlap * sampling
        flow = 5 * sampling / exp(log((chunk - overlap)/4., 2))
        assert (chunkp - overlapp) >= 2 * psdsize, (
            "Chunk duration not large enough to resolve lower-frequency "
            "bound, Omicron needs at least %ds. Minimum lower-frequency bound "
            "for this chunk duration is %.2gHz" % (2 * psdlen + overlap, flow))

    @integer_segments
    def output_segments(self, start, end):
        """Prints the list of processing segments for a given data segment
        """
        # get parameters
        if self.version >= 'v2r2':
            segment, overlap = self.getfloats('PARAMETER', 'TIMING')
            fileduration = segment - overlap
        else:
            chunk = self.getfloat('PARAMETER', 'CHUNKDURATION')
            segment = self.getfloat('PARAMETER', 'SEGMENTDURATION')
            overlap = self.getfloat('PARAMETER', 'OVERLAPDURATION')
            fileduration = chunk - overlap
        filesegment = segment - overlap
        padding = overlap / 2.

        # build list of file segments
        t = start + padding
        stop = end - padding
        segments = SegmentList()
        while t < stop:
            e = min(t + fileduration, stop)
            seg = Segment(t, e)
            if filesegment < abs(seg) < fileduration:
                remaining = abs(seg)
                nseg = remaining // filesegment * filesegment
                segments.append(Segment(t, t+nseg))
                if nseg != remaining:
                    segments.append(Segment(t+nseg, t+remaining))
            else:
                segments.append(seg)
            t = e
        return segments

    @integer_segments
    def distribute_segment(self, start, end, nperjob=1):
        """Determine processing segments to parallelise an Omicron job

        This function is meant to return a `segmentlist` of job [start, stop)
        times to pass to condor. Each segment will have duration
        `chunk * nperjob` *OR* `chunk * nperjob + remainder` if the remainder
        until the `end` is less than 1 chunk.

        Parameters
        ----------
        start : `int`
            the start GPS time of the Omicron segment

        end : `int`
            the end GPS time of the Omicron segment

        nperjob : `int`
            the number of chunks to put into each job

        Returns
        -------
        jobsegs : :class:`~ligo.segments.segmentlist`
            a `segmentlist` of [start, stop) times over which to distribute
            a single segment under condor
        """
        # get parameters
        if self.version >= 'v2r2':
            chunk = self.getfloat('PARAMETER', 'PSDLENGTH')
            _, overlap = self.getfloats('PARAMETER', 'TIMING')
        else:
            chunk = self.getfloat('PARAMETER', 'CHUNKDURATION')
            overlap = self.getfloat('PARAMETER', 'OVERLAPDURATION')

        # single small segment
        if end - start <= chunk * 2:
            return SegmentList([Segment(start, end)])

        # multiple larger segments
        out = SegmentList()
        t = start
        while t < end - overlap:
            seg = Segment(t, t)
            c = chunk
            while abs(seg) < chunk * nperjob and seg[1] < end:
                seg = Segment(seg[0], min(seg[1] + c, end))
                c = chunk - overlap
            if abs(seg) < chunk:
                out[-1] += seg
            else:
                out.append(seg)
            t = seg[1] - overlap
        return out

    def output_formats(self):
        return [fmt for fmt in ('root', 'txt', 'xml', 'hdf5') if
                fmt in self.get('OUTPUT', 'FORMAT')]

    def output_files(self, start, end, flatten=False):
        """Prints the list of output files for a given processing segment

        Parameters
        ----------
        start : `int`
            GPS start time of job

        end : `int`
            GPS end time of job

        flatten : `bool`, optional, default: `False`
            return a flat list of file names, rather than the default
            `dict` structure

        Returns
        -------
        files : `dict` of `dicts`
            `dict` of ``(channel, fdict)`` pairs, where ``fdict`` is a `dict`
            of ``(fileformat, filenames)`` pairs.
        """
        segments = self.output_segments(start, end)
        fileformats = self.output_formats()
        extension = {
            'root': 'root',
            'txt': 'txt',
            'xml': 'xml',
            'hdf5': 'h5',
        }

        # build list of files
        outdir = self.get('OUTPUT', 'DIRECTORY')
        out = {}
        for channel in self.getlist('DATA', 'CHANNELS'):
            cstr = CHANNEL_DELIM_REGEX.sub('_', channel).replace('_', '-', 1)
            out[channel] = dict((form, []) for form in fileformats)
            for seg in segments:
                basename = '%s_OMICRON-%d-%d' % (cstr, seg[0], abs(seg))
                for form in fileformats:
                    out[channel][form].append(os.path.join(
                        outdir, channel,
                        '{0}.{1}'.format(basename, extension[form]),
                    ))
        if flatten:  # return a basic list of filenames
            return [f for c in out for form in out[c] for f in out[c][form]]
        return out



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Test parameter handling for Omicron
"""

import os
import tempfile
try:
    from configparser import ConfigParser
except ImportError:  # python < 3
    from ConfigParser import ConfigParser

import pytest

from ..parameters import OmicronParameters

TEST_PARAMETERS = b"""
OUTPUT DIRECTORY /tmp/omicron-test
OUTPUT VERBOSTIY 1
DATA CHANNELS L1:CAL-DELTAL_EXTERNAL_DQ
DATA CHANNELS L1:CAL-DARM_CTRL_WHITEN_OUT_DBL_DQ
DATA CHANNELS L1:CAL-DARM_ERR_WHITEN_OUT_DBL_DQ
"""


def create(*args, **kwargs):
    return OmicronParameters(*args, **kwargs)


@pytest.fixture
def pars():
    return create(version='v2r2')


def test_validate_parameters(pars):
    # test zero overlap
    pars.set('PARAMETER', 'PSDLENGTH', '8')
    pars.set('PARAMETER', 'TIMING', '4 0')
    with pytest.raises(AssertionError) as excinfo:
        pars.validate()
    assert 'cannot run with zero overlap' in str(excinfo)

    # test odd-valued overlap
    pars.set('PARAMETER', 'TIMING', '4 1')
    with pytest.raises(AssertionError) as excinfo:
        pars.validate()
    assert 'Padding (overlap/2) is non-integer' in str(excinfo)

    # test odd-valued overlap
    pars.set('PARAMETER', 'TIMING', '8 6')
    with pytest.raises(AssertionError) as excinfo:
        pars.validate()
    assert 'Overlap is too large, cannot be more than 50%' in str(excinfo)

    # test odd-valued overlap
    pars.set('PARAMETER', 'TIMING', '4 5')
    with pytest.raises(AssertionError) as excinfo:
        pars.validate()
    assert 'Overlap length is greater than segment length' in str(excinfo)

    # test segment duration too long
    pars.set('PARAMETER', 'TIMING', '9 3')
    with pytest.raises(AssertionError) as excinfo:
        pars.validate()
    assert 'Segment length is greater than chunk length' in str(excinfo)

    # tests ahoy
    pars.set('PARAMETER', 'PSDLENGTH', '64')
    pars.set('PARAMETER', 'TIMING', '64 8')
    pars.validate()

    pars.set('PARAMETER', 'PSDLENGTH', '232')
    pars.validate()

    # test chunks and segments match
    pars.set('PARAMETER', 'PSDLENGTH', '128')
    pars.set('PARAMETER', 'TIMING', '65 8')
    with pytest.raises(AssertionError) as excinfo:
        pars.validate()
    assert 'Chunk duration doesn\'t allow an integer' in str(excinfo)


def test_from_channel_list_config():
    cp = ConfigParser()
    section = 'test'
    cp.add_section(section)
    cp.set(section, 'channels', 'X1:TEST-CHANNEL\nX1:TEST-CHANNEL_2')
    cp.set(section, 'flow', '10')
    cp.set(section, 'fhigh', '100')
    with tempfile.NamedTemporaryFile(suffix='.ini', mode='w') as f:
        cp.write(f)
        pars = OmicronParameters.from_channel_list_config(cp, section)
    assert pars.getlist('DATA', 'CHANNELS') == ['X1:TEST-CHANNEL',
                                                'X1:TEST-CHANNEL_2']
    assert tuple(pars.getfloats('PARAMETER', 'FREQUENCYRANGE')) == (10., 100.)


def test_read_ini(pars):
    cp = ConfigParser()
    section = 'DATA'
    cp.add_section(section)
    cp.set(section, 'channels', 'X1:TEST-CHANNEL')
    with tempfile.NamedTemporaryFile(suffix='.ini', mode='w') as f:
        cp.write(f)
        f.seek(0)
        pars.read(f.name)
    assert pars.getlist('DATA', 'CHANNELS') == ['X1:TEST-CHANNEL']


def test_read_txt(pars):
    with tempfile.NamedTemporaryFile(suffix='.txt') as f:
        f.write(TEST_PARAMETERS)
        f.seek(0)
        pars.readfp(f)
    assert pars.getlist('DATA', 'CHANNELS') == [
        'L1:CAL-DELTAL_EXTERNAL_DQ',
        'L1:CAL-DARM_CTRL_WHITEN_OUT_DBL_DQ',
        'L1:CAL-DARM_ERR_WHITEN_OUT_DBL_DQ',
    ]


def _test_write(pars, suffix):
    pars.set('DATA', 'CHANNELS',
             'L1:CAL-DELTAL_EXTERNAL_DQ '
             'L1:CAL-DARM_CTRL_WHITEN_OUT_DBL_DQ')
    pars.set('PARAMETER', 'PSDLENGTH', '124')
    pars.set('PARAMETER', 'TIMING', '64 4')
    with tempfile.NamedTemporaryFile(suffix=suffix, mode='w') as f:
        pars.write(f)
        f.seek(0)
        p2 = create()
        p2.read(f.name)
    assert pars.get('DATA', 'CHANNELS') == p2.get('DATA', 'CHANNELS')
    assert tuple(p2.getfloats('PARAMETER', 'TIMING')) == (64.0, 4.0)


def test_write_ini(pars):
    _test_write(pars, '.ini')


def test_write_txt(pars):
    _test_write(pars, '.txt')


def test_write_distributed(pars):
    stub = 'X1:TEST_CHANNEL_%d'
    channels = [stub % i for i in range(50)]
    pars.set('PARAMETER', 'PSDLENGTH', '100')
    pars.set('DATA', 'CHANNELS', ' '.join(channels))
    tmpdir = tempfile.gettempdir()
    try:
        _, files = pars.write_distributed(tmpdir, nchannels=10)
        for i, f in enumerate(files):
            cset = channels[i*10: (i+1)*10]
            p2 = create()
            p2.read(f)
            assert cset == p2.getlist('DATA', 'CHANNELS')
            assert p2.getfloat('PARAMETER', 'PSDLENGTH') == 100
    finally:
        for path in filter(os.path.isfile, files):
            os.remove(path)


def test_output_segments(pars):
    pars.set('PARAMETER', 'TIMING', '64 4')
    segs = pars.output_segments(0, 100)
    assert segs == [(2, 62), (62, 98)]


def test_distribute_segments(pars):
    pars.set('PARAMETER', 'TIMING', '64 4')
    pars.set('PARAMETER', 'PSDLENGTH', '124')
    segs = pars.distribute_segment(0, 1000, nperjob=4)
    segs == [(0, 604), (600, 1000)]


def test_output_files(pars):
    pars.set('PARAMETER', 'TIMING', '64 4')
    pars.set('DATA', 'CHANNELS', 'X1:TEST-CHANNEL')
    assert pars.output_files(0, 100) == {
        'X1:TEST-CHANNEL': {
             'root': [
                 './X1:TEST-CHANNEL/X1-TEST_CHANNEL_OMICRON-2-60.root',
                 './X1:TEST-CHANNEL/X1-TEST_CHANNEL_OMICRON-62-36.root',
             ],
             'xml': [
                 './X1:TEST-CHANNEL/X1-TEST_CHANNEL_OMICRON-2-60.xml',
                 './X1:TEST-CHANNEL/X1-TEST_CHANNEL_OMICRON-62-36.xml',
             ],
        },
    }



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Segment utilities for Omicron
"""

from __future__ import print_function

import json
import re
import warnings
from math import (floor, ceil)
from functools import wraps

from dqsegdb2.query import DEFAULT_SEGMENT_SERVER
from dqsegdb2.http import request as dqsegdb2_request

from gwpy.io.cache import (cache_segments as _cache_segments, file_segment)
from gwpy.io.gwf import data_segments as gwf_data_segments
from gwpy.segments import (DataQualityFlag, Segment, SegmentList)
from gwpy.timeseries import (StateTimeSeries, StateVector, TimeSeriesDict)
from gwpy.time import LIGOTimeGPS

from . import data

STATE_CHANNEL = {
    "H1:DMT-GRD_ISC_LOCK_NOMINAL:1": (
        "H1:GRD-ISC_LOCK",  # channel name or guardian node prefix
        "guardian",  # state bits (or 'guardian')
        "H1_R",  # state frametype
    ),
    "L1:DMT-GRD_ISC_LOCK_NOMINAL:1": (
        "L1:GRD-ISC_LOCK",
        "guardian",
        "L1_R",
    ),
    "H1:DMT-UP:1": (
        "H1:GDS-CALIB_STATE_VECTOR",
        [2],
        "H1_HOFT_C00",
    ),
    "L1:DMT-UP:1": (
        "L1:GDS-CALIB_STATE_VECTOR",
        [2],
        "L1_HOFT_C00",
    ),
    "H1:DMT-CALIBRATED:1": (
        "H1:GDS-CALIB_STATE_VECTOR",
        [0],
        "H1_HOFT_C00",
    ),
    "L1:DMT-CALIBRATED:1": (
        "L1:GDS-CALIB_STATE_VECTOR",
        [0],
        "L1_HOFT_C00",
    ),
    "H1:DMT-ANALYSIS_READY:1": (
        "H1:GDS-CALIB_STATE_VECTOR",
        [0, 1, 2],
        "H1_HOFT_C00",
    ),
    "L1:DMT-ANALYSIS_READY:1": (
        "L1:GDS-CALIB_STATE_VECTOR",
        [0, 1, 2],
        "L1_HOFT_C00",
    ),
    "V1:ITF_LOCKED:1": (
        "V1:DQ_ANALYSIS_STATE_VECTOR",
        [11],
        "V1_llhoft",
    ),
    "V1:ITF_SCIENCE:1": (
        "V1:DQ_ANALYSIS_STATE_VECTOR",
        [0, 1, 2],
        "V1_llhoft",
    ),
}
RAW_TYPE_REGEX = re.compile('[A-Z]1_R')


def integer_segments(f):
    @wraps(f)
    def decorated_method(*args, **kwargs):
        segs = f(*args, **kwargs)
        return type(segs)(type(s)(int(s[0]), int(s[1])) for s in segs)
    return decorated_method


def read_segments(filename, coltype=int):
    return SegmentList.read(
        filename,
        gpstype=coltype,
        format="segwizard",
    )


def get_last_run_segment(segfile):
    return read_segments(segfile, coltype=int)[-1]


def write_segments(segmentlist, outfile, coltype=int):
    return SegmentList(segmentlist).write(
        outfile,
        coltype=coltype,
        format="segwizard",
    )


@integer_segments
def query_state_segments(flag, start, end, url=DEFAULT_SEGMENT_SERVER,
                         pad=(0, 0)):
    """Query a segment database for active segments associated with a flag
    """
    # NOTE: DQF.pad pads forward in time at end
    return DataQualityFlag.query(
        flag, start-pad[0], end+pad[1], url=url,
    ).coalesce().pad(pad[0], -pad[1]).active


@integer_segments
def get_state_segments(channel, frametype, start, end, bits=[0], nproc=1,
                       pad=(0, 0)):
    """Read state segments from a state-vector channel in the frames
    """
    ifo = channel[:2]
    pstart = start - pad[0]
    pend = end + pad[1]

    # find frame cache
    cache = data.find_frames(ifo, frametype, pstart, pend)

    # optimise I/O based on type and library
    io_kw = {}
    try:
        from LDAStools import frameCPP  # noqa: F401
    except ImportError:
        pass
    else:
        io_kw['format'] = 'gwf.framecpp'
        if RAW_TYPE_REGEX.match(frametype):
            io_kw['type'] = 'adc'
        elif channel.endswith('GDS-CALIB_STATE_VECTOR'):
            io_kw['type'] = 'proc'

    bits = map(str, bits)
    # FIXME: need to read from cache with single segment but doesn't match
    # [start, end)

    # Virgo drops the state vector regularly, so need to sieve the files
    if channel == "V1:DQ_ANALYSIS_STATE_VECTOR":
        span = gwf_data_segments(cache, channel)
    else:
        span = SegmentList([Segment(pstart, pend)])

    # read data segments
    segs = SegmentList()
    try:
        csegs = cache_segments(cache)
    except KeyError:
        return segs
    for seg in csegs & span:
        sv = StateVector.read(cache, channel, nproc=nproc, start=seg[0],
                              end=seg[1], bits=bits, gap='pad', pad=0,
                              **io_kw).astype('uint32')
        segs += sv.to_dqflags().intersection().active

    # truncate to integers, and apply padding
    for i, seg in enumerate(segs):
        segs[i] = type(seg)(int(ceil(seg[0])) + pad[0],
                            int(floor(seg[1])) - pad[1])
    segs.coalesce()

    return segs.coalesce()


@integer_segments
def get_frame_segments(obs, frametype, start, end):
    cache = data.find_frames(obs, frametype, start, end)
    span = SegmentList([Segment(start, end)])
    return cache_segments(cache) & span


@integer_segments
def get_guardian_segments(node, frametype, start, end, nproc=1, pad=(0, 0),
                          strict=False):
    """Determine state segments for a given guardian node
    """
    ifo, node = node.split(':', 1)
    if node.startswith('GRD-'):
        node = node[4:]
    pstart = start - pad[0]
    pend = end + pad[1]

    # find frame cache
    cache = data.find_frames(ifo, frametype, pstart, pend)

    # pre-format data segments
    span = SegmentList([Segment(pstart, pend)])
    segs = SegmentList()
    csegs = cache_segments(cache)
    if not csegs:
        return csegs

    # read data
    stub = "{}:GRD-{}".format(ifo, node)
    if strict:
        channels = ["{}_OK".format(stub)]
    else:
        state = "{}_STATE_N".format(stub)
        nominal = "{}_NOMINAL_N".format(stub)
        active = "{}_ACTIVE".format(stub)
        channels = [state, nominal, active]
    for seg in csegs & span:
        if strict:
            sv = StateVector.read(
                cache, channels[0], nproc=nproc, start=seg[0], end=seg[1],
                bits=[0], gap='pad', pad=0,).astype('uint32')
            segs += sv.to_dqflags().intersection().active
        else:
            gdata = TimeSeriesDict.read(
                cache, channels, nproc=nproc, start=seg[0], end=seg[1],
                gap='pad', pad=0)
            ok = ((gdata[state].value == gdata[nominal].value) &
                  (gdata[active].value == 1)).view(StateTimeSeries)
            ok.t0 = gdata[state].t0
            ok.dt = gdata[state].dt
            segs += ok.to_dqflag().active

    # truncate to integers, and apply padding
    for i, seg in enumerate(segs):
        segs[i] = type(seg)(int(ceil(seg[0])) + pad[0],
                            int(floor(seg[1])) - pad[1])
    segs.coalesce()

    return segs.coalesce()


@integer_segments
def cache_segments(cache):
    return _cache_segments(cache).coalesce()


def segmentlist_from_tree(tree, coalesce=False):
    """Read a `~ligo.segments.segmentlist` from a 'segments' `ROOT.Tree`
    """
    segs = SegmentList()
    for i in range(tree.GetEntries()):
        tree.GetEntry(i)
        segs.append(Segment(tree.start, tree.end))
    return segs


def get_flag_coverage(flag, url=DEFAULT_SEGMENT_SERVER):
    """Return the coverage data for the given flag
    """
    ifo, name, version = flag.rsplit(':', 2)
    flagu = '/dq/%s/%s/%s' % (ifo, name, version)
    raw = dqsegdb2_request('%s/report/coverage' % url)
    return json.loads(raw.read().decode('utf-8'))['results'][flagu]


def get_latest_active_gps(flag, url=DEFAULT_SEGMENT_SERVER):
    """Return the end time of the latest active segment for this flag
    """
    return get_flag_coverage(flag, url=url)['latest_active_segment']


def get_latest_known_gps(flag, url=DEFAULT_SEGMENT_SERVER):
    """Return the end time of the latest known segment for this flag
    """
    return get_flag_coverage(flag, url=url)['latest_known_segment']


@integer_segments
def cache_overlaps(*caches):
    """Find segments of overlap in the given cache sets
    """
    cache = [e for c in caches for e in c]
    cache.sort(key=lambda e: file_segment(e)[0])
    overlap = SegmentList()
    segments = SegmentList()
    for e in cache:
        seg = file_segment(e)
        ol = SegmentList([seg]) & segments
        if abs(ol):
            overlap.extend(ol)
        segments.append(seg)
    return overlap



#!/usr/bin/env python

"""This module is a mock of htcondor.py, purely for testing
"""


class Schedd(object):
    """Mock of the `htcondor.Schedd` object
    """
    _jobs = []

    def query(self, constraints, attr_list=[], **kwargs):
        print('_jobs', self._jobs)
        x = {}
        for con in constraints.split(' && '):
            try:
                a, b = con.split(' == ')
            except ValueError:
                break
            else:
                x[a] = eval(b)
        match = []

        def match(job):
            for key in x:
                if x[key] != job[key]:
                    return False
            return True

        return iter(filter(match, self._jobs))

    def history(self, *args, **kwargs):
        return



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2019)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for omicron.segments
"""

import tempfile
from copy import deepcopy
try:
    from unittest import mock
    from io import StringIO
except ImportError:  # python < 3
    from StringIO import StringIO
    import mock

import pytest

from gwpy.segments import (DataQualityFlag, Segment, SegmentList)

from .. import segments


@pytest.fixture
def seglist():
    return SegmentList([
        Segment(0, 1),
        Segment(1, 2),
        Segment(3, 4),
    ])


def test_read_write_segments(seglist):
    with tempfile.NamedTemporaryFile(mode="w") as tmp:
        segments.write_segments(seglist, tmp)
        tmp.seek(0)
        segs = segments.read_segments(tmp.name)
        assert segs == seglist


def test_get_last_run_segment(seglist):
    tmp = StringIO()
    segments.write_segments(seglist, tmp)
    tmp.seek(0)
    assert segments.get_last_run_segment(tmp) == seglist[-1]


def test_query_state_segments(seglist):
    with mock.patch(
        "omicron.segments.DataQualityFlag.query",
        return_value=DataQualityFlag(active=deepcopy(seglist),
                                     known=[seglist.extent()]),
    ):
        coal = deepcopy(seglist).coalesce()
        assert segments.query_state_segments('X', 0, 10) == coal
        assert segments.query_state_segments(
            'X', 0, 10,
            pad=(1, 1),
        ) == DataQualityFlag(active=coal,
                             known=[coal.extent()]).pad(1, -1).active


@mock.patch(
    "omicron.data.find_frames",
    return_value=["/path/to/A-B-0-10.gwf", "/path/to/C-D-20-10.gwf"],
)
def test_get_frame_segments(find):
    assert segments.get_frame_segments(
        "X",
        "X1_R",
        0,
        100
    ) == SegmentList([Segment(0, 10), Segment(20, 30)])
    assert segments.get_frame_segments(
        "X",
        "X1_R",
        25,
        100,
    ) == SegmentList([Segment(25, 30)])



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Condor interaction utilities
"""

from __future__ import print_function

import os.path
import re
import time
import warnings
from datetime import datetime
from time import sleep
from os import stat
from glob import glob
from getpass import getuser
from subprocess import CalledProcessError

import htcondor
from classad import ClassAd

import numpy

from glue import pipeline

from gwpy.time import to_gps

from .utils import (shell, which)

re_dagman_cluster = re.compile(r'(?<=submitted\sto\scluster )[0-9]+')

JOB_STATUS = [
    'Unexpanded',
    'Idle',
    'Running',
    'Removed',
    'Completed',
    'Held',
    'Submission error',
]
JOB_STATUS_MAP = dict((v.lower(), k) for k, v in enumerate(JOB_STATUS))


JOB_UNIVERSE = [
    'Min',
    'Standard',
    'Pipe',
    'Linda',
    'PVM',
    'Vanilla',
    'PVMD',
    'Scheduler',
    'MPI',
    'Grid',
    'Java',
    'Parallel',
    'Local',
    'Max',
]


def submit_dag(dagfile, *arguments, **options):
    """Submit a DAG to condor and return the cluster ID

    Parameters
    ----------
    dagfile : `str`
        path to DAG file for submission
    *arguments
        other command-line arguments to pass directly to condor_submit_dag
    **options
        other `(key, value)` pairs to append to the condor_submit_dag
        command

    Returns
    -------
    cluster : `int`
        the cluster ID for the newly submitted DAG

    Raises
    ------
    subprocess.CalledProcessError
        if the call to `condor_submit_dag` fails for some reason
    """
    cmd = [which('condor_submit_dag')] + list(arguments)
    for opt, val in options.items():
        cmd.extend([opt, val])
    cmd.append(dagfile)
    print("$ %s" % ' '.join(cmd))
    out = shell(cmd)
    print(out)
    try:
        return int(re_dagman_cluster.search(out).group())
    except (AttributeError, IndexError, TypeError) as e:
        e.args = ('Failed to extract DAG cluster ID from '
                  'condor_submit_dag output [%s]' % str(e),)
        raise


def monitor_dag(dagfile, interval=5):
    """Monitor the status of a DAG by watching the .lock file
    """
    lock = '%s.lock' % dagfile
    stat(lock)
    while True:
        sleep(interval)
        try:
            stat(lock)
        except OSError:
            break
    try:
        find_rescue_dag(dagfile)
    except IndexError:
        return


def find_rescue_dag(dagfile):
    """Find the most recent rescue DAG related to this DAG

    Returns
    -------
    rescue : `str`
        the path to the highest-enumerated rescue DAG

    Raises
    ------
    IndexError
        if no related rescue DAG files are found
    """
    try:
        return sorted(glob('%s.rescue[0-9][0-9][0-9]' % dagfile))[-1]
    except IndexError as e:
        e.args = ('No rescue DAG files found',)
        raise


def iterate_dag_status(clusterid, interval=2):
    """Monitor a DAG by querying condor for status information periodically
    """
    schedd = htcondor.Schedd()
    while True:
        try:
            status = get_dag_status(clusterid, schedd=schedd, detailed=True)
        except (IOError, KeyError) as e:
            # reconnect, and try again
            sleep(1)
            del schedd
            schedd = htcondor.Schedd()
            try:
                status = get_dag_status(clusterid, schedd=schedd,
                                        detailed=True)
            except IOError:
                raise e
        yield status
        if 'exitcode' in status:
            break
        sleep(interval)


def get_dag_status(dagmanid, schedd=None, detailed=True):
    """Return the status of a given DAG

    Parameters
    ----------
    dagmanid : `int`
        the ClusterId of the DAG
    schedd : `htcondor.Schedd`, optional
        the open connection to the scheduler
    detailed : `bool`, optional
        check jobs as held

    Returns
    -------
    status : `dict`
        a `dict` summarising the DAG status with the following keys

        - 'total': the total number of jobs
        - 'done': the number of completed jobs
        - 'queued': the number of queued jobs (excluding held if `held=True`)
        - 'ready': the number of jobs ready to be submitted
        - 'unready': the number of jobs not ready to be submitted
        - 'failed': the number of failed jobs
        - 'held': the number of failed jobs (only non-zero if `held=True`)

        Iff the DAG is completed, the 'exitcode' of the DAG will be included
        in the returned status `dict`
    """
    # connect to scheduler
    if schedd is None:
        schedd = htcondor.Schedd()
    # find running DAG job
    states = ['total', 'done', 'queued', 'ready', 'unready', 'failed']
    classads = ['DAG_Nodes%s' % s.title() for s in states]
    try:
        job = find_job(ClusterId=dagmanid, schedd=schedd, attr_list=classads)
        status = dict()
        for s, c in zip(states, classads):
            try:
                status[s] = job[c]
            except KeyError:  # htcondor.py failure (unknown cause)
                try:
                    status[s] = int(shell(['condor_q', str(dagmanid),
                                           '-autoformat', c]))
                except (ValueError, CalledProcessError):
                    status[s] = '-'
    # DAG has exited
    except RuntimeError as e:
        if not str(e).startswith('No jobs found'):
            raise
        sleep(1)
        try:
            job = list(schedd.history('ClusterId == %d' % dagmanid,
                                      classads+['ExitCode'], 1))[0]
        except IOError:  # try again
            job = list(schedd.history('ClusterId == %d' % dagmanid,
                                      classads+['ExitCode'], 1))[0]
        except KeyError:  # condor_rm not finished yet (probably)
            sleep(10)
            job = list(schedd.history('ClusterId == %d' % dagmanid,
                                      classads+['ExitCode'], 1))[0]
        except RuntimeError as e:
            if 'timeout' in str(e).lower() or 'cowardly' in str(e).lower():
                job = get_condor_history_shell(
                    'ClusterId == %d' % dagmanid,
                    classads+['ExitCode'], 1)[0]
                job = dict((k, int(v)) for k, v in job.items())
            else:
                raise
        history = dict((s, job[c]) for s, c in zip(states, classads))
        history['exitcode'] = job['ExitCode']
        history['held'] = history['running'] = history['idle'] = 0
        return history
    # DAG is running, get status
    else:
        # find node status details
        if detailed:
            status['held'] = 0
            status['running'] = 0
            status['idle'] = 0
            nodes = find_jobs(DAGManJobId=dagmanid, schedd=schedd)
            for node in nodes:
                s = get_job_status(node)
                if s == JOB_STATUS_MAP['held']:
                    status['held'] += 1
                elif s == JOB_STATUS_MAP['running']:
                    status['running'] += 1
                elif s == JOB_STATUS_MAP['idle']:
                    status['idle'] += 1
        return status


def get_job_duration_history_shell(classad, value, user=getuser(),
                                   maxjobs=None):
    """Return the durations of history condor jobs

    This method calls to `condor_history` in the shell.

    Parameters
    ----------
    classad : `str`
        name of classad providing unique identifier for job type
    value :
        value of classad
    user : `str`, optional
        name of submitting user
    maxjobs : `int`, optional
        maximum number of matches to return

    Returns
    -------
    times, durations : `tuple` of `numpy.ndarray`
        two arrays with the job end time and durations of each matched
        condor process
    """
    if isinstance(value, str):
        value = '"%s"' % value
    cmd = ['condor_history', '-constraint',
           '\'%s==%s && Owner=="%s"\'' % (classad, value, user),
           '-autof', 'EnteredCurrentStatus',
           '-autof', 'JobStartDate']
    if maxjobs is not None:
        cmd.extend(['-match', str(maxjobs)])
    history = shell(' '.join(cmd), shell=True)
    lines = history.rstrip('\n').split('\n')
    times = numpy.zeros(len(lines))
    jobdur = numpy.zeros(times.size)
    for i, line in enumerate(lines):
        try:
            e, s = map(int, line.split())
        except ValueError:
            times = times[:i]
            jobdur = jobdur[:i]
            break
        times[i] = to_gps(datetime.fromtimestamp(e)) + time.timezone
        jobdur[i] = e - s
    return times, jobdur


def get_job_duration_history(classad, value, user=getuser(), maxjobs=0,
                             schedd=None):
    """Return the durations of history condor jobs

    This method uses the python bindings for `htcondor`, which seems
    to have network transfer limits, do not use for large job numbers
    (>2000), instead use `get_job_duration_history_shell` which calls
    to `condor_history` in the shell.

    Parameters
    ----------
    classad : `str`
        name of classad providing unique identifier for job type
    value :
        value of classad
    user : `str`, optional
        name of submitting user
    maxjobs : `int`, optional
        maximum number of matches to return

    Returns
    -------
    times, durations : `tuple` of `numpy.ndarray`
        two arrays with the job end time and durations of each matched
        condor process
    """
    if schedd is None:
        schedd = htcondor.Schedd()
    if isinstance(value, str):
        value = '"%s"' % value
    history = list(schedd.history(
        '%s==%s && Owner=="%s"' % (classad, value, user),
        ['EnteredCurrentStatus', 'JobStartDate'], maxjobs))
    times = numpy.zeros(len(history))
    jobdur = numpy.zeros(len(history))
    for i, h in enumerate(history):
        times[i] = (
            to_gps(datetime.fromtimestamp(h['EnteredCurrentStatus'])) +
            time.timezone)
        jobdur[i] = h['EnteredCurrentStatus'] - h['JobStartDate']
    return times, jobdur


def get_condor_history_shell(constraint, classads, maxjobs=None):
    """Get condor_history from the shell

    Parameters
    ----------
    constraint : `str`
        `str` of the format 'ClassAd == "value"' defining the `-constraint`
        to pass to `condor_history`
    classads : `list` of `str`
        list of class Ad names to get back from `condor_history`
    maxjobs : `int`
        the number of matches to return

    Returns
    -------
    jobs : `list` of `dict`
        list of dicts with same keys as defined by `get_dag_status`
    """
    cmd = ['condor_history', '-constraint', constraint]
    for ad_ in classads:
        cmd.extend(['-autof', ad_])
    if maxjobs:
        cmd.extend(['-match', str(maxjobs)])
    history = shell(' '.join(cmd), shell=True)
    lines = history.rstrip('\n').split('\n')
    jobs = []
    for line in lines:
        values = line.split()
        jobs.append(dict(zip(classads, values)))
    return jobs


def get_out_err_files(dagmanid, exitcode=None, schedd=None, user=getuser(),
                      maxjobs=0):
    """Get the paths of the output and error files for nodes in a given DAG

    Parameters
    ----------
    dagmanid : `int`
        the ClusterId of the DAG
    exitcode : `int`, optional
        return only nodes with this exitcode, or return all nodes if
        `None`
    schedd : `htcondor.Schedd`, optional
        the open connection to the scheduler
    user : `str`, optional
        the name of the user who submitted the DAG, defaults to you
    maxjobs : `int`, optional
        maximum number of condor history records to return, defaults
        to `0` meaning 'all'

    Returns
    -------
    filedict : `dict`
        a `dict` of `(nodeid, [files])` pairs
    """
    if schedd is None:
        schedd = htcondor.Schedd()
    history = list(schedd.history(
        'DAGManJobId==%d && Owner=="%s"' % (dagmanid, user),
        ['ExitCode', 'Out', 'Err', 'ClusterId'], maxjobs))
    out = {}
    for node in history:
        try:
            ec = node['ExitCode']
        except KeyError:
            warnings.warn("Failed to get ExitCode for node %r" % node)
            continue
        if exitcode is not None and ec != exitcode:
            continue
        out[node['ClusterId']] = [node['Out'], node['Err']]
    return out


def find_dagman_id(group, classad="OmicronDAGMan", user=getuser(),
                   schedd=None):
    """Find the condor ID of the DAGMan process for the given group

    Parameters
    ----------
    group : `str`
        name of group to find, actually this is just the value of the classad

    classad : `str`, optional
        the defining classad used to identify the correct DAGMan process

    user : `str`, optional
        the name of the submitting user, defaults to the current user

    schedd : `htcondor.Schedd`, optional
        open connection to the scheduler, one will be created if needed

    Returns
    -------
    clusterid : `int`
        the integer ID of the DAGMan process cluster

    Raises
    ------
    RuntimeError
        if not exactly 1 matching condor process is found, or that process
        is not in a good state
    """
    constraints = {classad: group}
    if user is not None:
        constraints['Owner'] = user
    job = find_job(schedd=schedd, attr_list=['JobStatus', 'ClusterId'],
                   **constraints)
    clusterid = job['ClusterId']
    if get_job_status(job) >= 3:
        raise RuntimeError("DAGMan cluster %d found, but in state %r"
                           % JOB_STATUS[job['JobStatus']])
    return clusterid


def dag_is_running(dagfile, schedd=None):
    """Return whether a DAG is running

    This method will return `True` if any of the following match

    - {dagfile}.lock file is found
    - a job is found in the condor queue with the UserLog classad matching
      {dagfile}.dagman.log

    Otherwise, the return is `False`

    Parameters
    ----------
    dagfile : `str`
        the path of the DAG you want to analyse

    Raises
    ------
    RuntimeError
        if multiple matching condor processes are found, or the matching
        single process is not in a good state
    """
    if os.path.isfile('%s.lock' % dagfile):
        return True
    userlog = '%s.dagman.log' % dagfile
    try:
        find_job(UserLog=userlog, schedd=schedd)
    except RuntimeError as e:
        if str(e).startswith('No jobs found'):
            return False
        raise
    else:
        return True
    return False


def get_job_status(job, schedd=None):
    """Get the status of a HTCondor process

    Parameters
    ----------
    job : `str`, `classad.ClassAd`
        either job ID (`str`, `int`, `float`) or a job object

    schedd : `htcondor.Schedd`, optional
        open scheduler connection

    Returns
    -------
    status : `int`
        the integer (`long`) status code for this job
    """
    if not isinstance(job, ClassAd):
        job = find_job(ClusterId=job, schedd=schedd, attr_list=['JobStatus'])
    return job['JobStatus']


def find_jobs(schedd=None, attr_list=None, **constraints):
    """Query the condor queue for jobs matching the constraints

    Parameters
    ----------
    schedd : `htcondor.Schedd`, optional
        open scheduler connection

    attr_list : `list` of `str`
        list of attributes to return for each job, defaults to all

    all other keyword arguments should be ClassAd == value constraints to
    apply to the scheduler query

    Returns
    -------
    jobs : `list` of `classad.ClassAd`
        the job listing for each job found
    """
    if schedd is None:
        schedd = htcondor.Schedd()
    qstr = ' && '.join(['%s == %r' % (k, v) for
                        k, v in constraints.items()]).replace("'", '"')
    if not attr_list:
        attr_list = []
    return list(schedd.query(qstr, attr_list))


def find_job(schedd=None, attr_list=None, **constraints):
    """Query the condor queue for a single job matching the constraints

    Parameters
    ----------
    schedd : `htcondor.Schedd`, optional
        open scheduler connection

    attr_list : `list` of `str`
        list of attributes to return for each job, defaults to all

    all other keyword arguments should be ClassAd == value constraints to
    apply to the scheduler query

    Returns
    -------
    classad : `classad.ClassAd`
        the job listing for the found job

    Raises
    ------
    RuntimeError
        if not exactly one job is found matching the constraints
    """
    jobs = find_jobs(schedd=schedd, attr_list=attr_list, **constraints)
    if len(jobs) == 0:
        raise RuntimeError("No jobs found matching constraints %r"
                           % constraints)
    elif len(jobs) > 1:
        raise RuntimeError("Multiple jobs found matching constraints %r"
                           % constraints)
    return jobs[0]


# -- custom jobs --------------------------------------------------------------

class OmicronProcessJob(pipeline.CondorDAGJob):
    """`~glue.pipe.CondorJob` as part of Omicron processing
    """
    logtag = '$(cluster)-$(process)'

    def __init__(self, universe, executable, tag=None, subdir=None,
                 logdir=None, request_memory=1000, **cmds):
        pipeline.CondorDAGJob.__init__(self, universe, executable)
        if tag is None:
            tag = os.path.basename(os.path.splitext(executable)[0])
        if subdir:
            subdir = os.path.abspath(subdir)
            self.set_sub_file(os.path.join(subdir, '%s.sub' % (tag)))
        if logdir:
            logdir = os.path.abspath(logdir)
            self.set_log_file(os.path.join(
                logdir, '%s-%s.log' % (tag, self.logtag)))
            self.set_stderr_file(os.path.join(
                logdir, '%s-%s.err' % (tag, self.logtag)))
            self.set_stdout_file(os.path.join(
                logdir, '%s-%s.out' % (tag, self.logtag)))
        cmds.setdefault('getenv', 'True')
        for key, val in cmds.items():
            if hasattr(self, 'set_%s' % key.lower()):
                getattr(self, 'set_%s' % key.lower())(val)
            else:
                self.add_condor_cmd(key, val)
        # add sub-command option
        self._command = None

    def add_opt(self, opt, value=''):
        pipeline.CondorDAGJob.add_opt(self, opt, str(value))
    add_opt.__doc__ = pipeline.CondorDAGJob.add_opt.__doc__

    def set_command(self, command):
        self._command = command

    def get_command(self):
        return self._command

    def write_sub_file(self):
        pipeline.CondorDAGJob.write_sub_file(self)
        if self.get_command():
            with open(self.get_sub_file(), 'r') as f:
                sub = f.read()
            sub = sub.replace('arguments = "', 'arguments = " %s'
                              % self.get_command())
            with open(self.get_sub_file(), 'w') as f:
                f.write(sub)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for omicron.io
"""

import sys
import os.path
import tempfile

try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest

# mock up condor modules for testing purposes
# (so we don't need to have htcondor installed)
from . import (mock_htcondor, mock_classad)
sys.modules['htcondor'] = mock_htcondor  # noqa
sys.modules['classad'] = mock_classad  # noqa

from .. import condor


# -- mock utilities -----------------------------------------------------------

def mock_shell_factory(output):
    def shell(*args, **kwargs):
        return output
    return shell


def mock_schedd_factory(jobs):
    class Schedd(mock_htcondor.Schedd):
        _jobs = jobs
    return Schedd


# -- tests --------------------------------------------------------------------

@mock.patch('omicron.condor.which', return_value=sys.executable)
@mock.patch('omicron.condor.shell',
            return_value='1 job(s) submitted to cluster 12345')
def test_submit_dag(shell, which):
    dagid = condor.submit_dag('test.dag')
    assert dagid == 12345


@mock.patch('omicron.condor.which', return_value=sys.executable)
@mock.patch('omicron.condor.shell',
            return_value='1 job(s) submitted to cluster 12345')
def test_submit_dag_append(shell, which):
    condor.submit_dag('test.dag', '-append', '+OmicronDAGMan="GW"')
    assert shell.called_with([sys.executable, '-append',
                              '+OmicronDAGMan="GW"', 'test.dag'])


@mock.patch('omicron.condor.which', return_value=sys.executable)
@mock.patch('omicron.condor.shell', return_value='Something else')
def test_submit_dag_error(shell, which):
    with pytest.raises(AttributeError) as exc:
        condor.submit_dag('test.dag')
    assert str(exc.value).startswith('Failed to extract DAG cluster ID')


@mock.patch('htcondor.Schedd',
            mock_schedd_factory([{'ClusterId': 1}, {'ClusterId': 2}]))
@pytest.mark.parametrize('kwargs, output', [
    ({}, [{'ClusterId': 1}, {'ClusterId': 2}]),
    ({'ClusterId': 1}, [{'ClusterId': 1}]),
    ({'ClusterId': 3}, []),
])
def test_find_jobs(kwargs, output):
    print('kwargs', kwargs)
    print('jobs', condor.find_jobs(**kwargs))
    print('output', output)
    assert condor.find_jobs(**kwargs) == output


@mock.patch('htcondor.Schedd',
            mock_schedd_factory([{'ClusterId': 1}, {'ClusterId': 2}]))
def test_find_job():
    job = condor.find_job(ClusterId=1)
    assert job == {'ClusterId': 1}

    # check 0 jobs returned throws the right error
    with pytest.raises(RuntimeError) as exc:
        condor.find_job(ClusterId=3)
    assert str(exc.value).startswith('No jobs found')

    # check multiple jobs returned throws the right error
    with pytest.raises(RuntimeError) as exc:
        condor.find_job()
    assert str(exc.value).startswith('Multiple jobs found')


@mock.patch('htcondor.Schedd',
            mock_schedd_factory([{'ClusterId': 1, 'JobStatus': 4}]))
def test_get_job_status():
    assert condor.get_job_status(1) == 4


@mock.patch('os.path.isfile')
def test_dag_is_running(isfile):
    isfile.return_value = False
    assert not condor.dag_is_running('test.dag')

    isfile.return_value = True
    assert condor.dag_is_running('test.dag')

    isfile.return_value = False
    schedd_ = mock_schedd_factory(
        [{'UserLog': 'test.dag.dagman.log'}])
    with mock.patch('htcondor.Schedd', schedd_):
        assert condor.dag_is_running('test.dag')
        assert not condor.dag_is_running('test2.dag')


def test_find_rescue_dag():
    with tempfile.NamedTemporaryFile(suffix='.dag.rescue001') as f:
        dagf = os.path.splitext(f.name)[0]
        assert condor.find_rescue_dag(dagf) == f.name
    with pytest.raises(IndexError) as exc:
        assert condor.find_rescue_dag(dagf)
    assert 'No rescue DAG files found' in str(exc.value)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Constants and variables for Omicron processing
"""

import os

from ligo.segments import segment as Segment

# -- generic parameters
try:
    IFO = os.environ['IFO']
except KeyError:
    from socket import getfqdn
    fqdn = getfqdn()
    if '.uni-hannover.' in fqdn:
        IFO = 'G1'
    elif '.ligo-wa.' in fqdn:
        IFO = 'H1'
    elif '.ligo-la.' in fqdn:
        IFO = 'L1'
    elif '.virgo.' in fqdn or '.ego-gw.' in fqdn:
        IFO = 'V1'
    else:
        IFO = None
    ifo = os.getenv('ifo')
else:
    ifo = os.getenv('ifo', IFO.lower())
SITE = os.getenv('SITE')
site = os.getenv('site', SITE and SITE.lower() or None)

# -- omicron directories
HOME = os.path.expanduser('~')
# where Omicron runs
OMICRON_BASE = os.path.join(HOME, 'omicron')
# where Omicron triggers are produced
OMICRON_PROD = os.path.join(OMICRON_BASE, 'online')
# archive storage directory
OMICRON_ARCHIVE = os.path.join(HOME, 'triggers')
# tag Omicron itself places on XML files
OMICRON_FILETAG = 'Omicron'

# omicron production version
OMICRON_VERSION = 'v2r1'

# omicron channel files
if ifo is not None:
    OMICRON_GROUP_FILE = os.path.join(OMICRON_PROD, '%s-groups.txt' % ifo)
    OMICRON_CHANNELS_FILE = os.path.join(OMICRON_PROD, '%s-channels.txt' % ifo)
else:
    OMICRON_GROUP_FILE = None
    OMICRON_CHANNELS_FILE = None

# epochs
EPOCH = {
    "VSR1": Segment(863557214, 875250014),
    "VSR2": Segment(931035615, 947100000),
    "S6c": Segment(947100001, 965599214),
    "VSR3": Segment(965599215, 971686015),
    "VSR4": Segment(991170015, 999061215),
    "ER2": Segment(1026086416, 1028332816),
    "ER3": Segment(1042088795, 1046510000),
    "POSTER3": Segment(1046510000, 1057795216),
    "ER4": Segment(1057795216, 1063000000),
    "POSTER4": Segment(1063000000, 1073779216),
    "ER5": Segment(1073779216, 1079049616),
    "POSTER5": Segment(1073779216, 1101700000),
    "ER6": Segment(1101700000, 1103700000),
    "POSTER6": Segment(1103700000, 1116930000),
    "ER7": Segment(1116930000, 1118700000),
    "POSTER7": Segment(1118700000, 1123804817),
    "ER8": Segment(1123804817, 1126137617),
    "O1": Segment(1126137617, 1134604817),
    "POSTO1": Segment(1134604817, 1500000000),
}


def epoch_at_gps(gps):
    """Return the named epoch that contains the given GPS time

    Parameters
    ----------
    gps : `int`
        the GPS time in question

    Returns
    -------
    epoch : `str`
        the name of the containing epoch

    Examples
    --------
    >>> epoch_at_gps(1126259462)
    'O1'

    Raises
    ------
    ValueError
        if the given `gps` isn't covered by any known epoch
    """
    for epoch in EPOCH:
        if gps in EPOCH[epoch]:
            return epoch
    raise ValueError("GPS time %d not covered by any known epoch" % gps)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Test logging for Omicron
"""

import re

from .. import log

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_bold():
    assert log.bold('TEST') == '\x1b[1mTEST\x1b[0m'


def test_color_text():
    assert log.color_text('TEST', 'blue') == '\x1b[1;34mTEST\x1b[0m'


def test_logger():
    # create logger
    logger = log.Logger('TEST')

    # fudge a record
    record = logger.makeRecord(logger.name, log.logging.DEBUG, 'FILE', 0,
                               'test message', (), None, 'FUNC', None)

    # test that the formatter prints the correct thing
    outhandler = logger.handlers[0]
    assert re.match(r'\[\x1b\[1mTEST\x1b\[0m \d+\]    '
                    r'\x1b\[1;34mDEBUG\x1b\[0m: test message',
                    outhandler.format(record))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Nagios utilities for PyOmicron
"""

from __future__ import (print_function, division)

import json as jsonlib
import sys
from getpass import getuser

import htcondor

from gwpy.io.cache import file_segment
from gwpy.time import to_gps

from . import (const, condor)
from .io import find_latest_omicron_file
from .data import get_latest_data_gps
from .segments import get_latest_active_gps

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def nagios_exit_factory(name, author=None, json=False, **kwargs):
    """Create a `nagios_exit` method for a given nagios monitor

    The `nagios_exit` method takes an exitcode (`int`), and a message and
    displays a reponse that is parseable by nagios, either via ntpe or
    the LIGO nagios JSON parser.
    """
    def nagios_exit(exitcode, message, timeout=1800):
        """Exit this program in a nagios-compatible manner

        Parameters
        ----------
        exitcode : `int`
            the exitcode for the nagios message, an integer between 0--3
            (inclusive)
        message : `str`
            the message to print to the screen
        """
        if json:
            out = {
                'created_gps': int(to_gps('now')),
                'status_intervals': [
                    {'start_sec': 0,
                     'end_sec': timeout,
                     'num_status': exitcode,
                     'txt_status': message},
                    {'start_sec': timeout,
                     'num_status': 3,
                     'txt_status': '%s not running' % name},
                ],
            }
            for key in kwargs:
                out[key] = kwargs[key]
            if author is not None:
                author_, email = author.rsplit(' ', 1)
                email = email.strip('<').rstrip('>')
                out['author'] = {'name': author_, 'email': email}
            if isinstance(json, str):
                with open(json, 'w') as f:
                    print(jsonlib.dumps(out), file=f)
            else:
                print(jsonlib.dumps(out))
        else:
            print(message)
        sys.exit(exitcode)
    return nagios_exit


def monitor_condor_status(group, user=getuser()):
    """Check the status of an Omicron Online group process
    """
    schedd = htcondor.Schedd()
    jobs = schedd.query('Owner == "%s" && OmicronManager == "%s"'
                        % (user, group))
    if len(jobs) == 0:
        raise IndexError(
            'No condor processes associated with group %r' % group,)
    elif len(jobs) > 1:
        raise IndexError(
            'Multiple OmicronOnline managers found matching Owner == %r && '
            'OmicronManager == %r' % (user, group))
    try:
        job = dict(jobs[0])
    except ValueError as e:
        e.args = ('Failed to parse condor classads for group %r as dict'
                  % group,)
        raise
    # get status
    try:
        return condor.JOB_STATUS[job['JobStatus']]
    except IndexError as e:
        e.args = ('JobStatus for %d unrecognised' % job['ClusterId'],)


def monitor_dag_status(group, user=getuser()):
    """Check the status of a DAG of Omicron processes

    Parameters
    ----------
    group : `str`
        the name of Omicron channel group to use

    Returns
    -------
    nagios_status : (`int`, `str`)
        an `(exitcode, message)` `tuple` to use as the nagios service status
        for this Omicron DAG
    """
    # find DAGMan process
    schedd = htcondor.Schedd()
    jobs = schedd.query('Owner == "%s" && OmicronDAGMan == "%s"'
                        % (user, group))
    if not len(jobs) == 1:
        return (2, "Not exactly one jobs matching OmicronProcess "
                   "== %r found (%d found)" % (group, len(jobs)))
    # get DAG id and get status
    dagman = dict(jobs[0])
    dagmanid = dagman['ClusterId']
    status = condor.get_dag_status(dagmanid, schedd=schedd, detailed=True)
    if status.get('exitcode', None) == 0:
        return (0, "DAG has completed successfully")
    elif status.get('exitcode', 0) > 0:
        return (2, "DAG has completed with exitcode %d" % status['exitcode'])
    elif status.get('held', 0) > 0:
        return (1, "DAG has %d held jobs" % status['held'])
    elif status['failed'] > 0:
        return (1, "DAG has %d failed jobs" % status['failed'])
    return (0, 'DAG is running [%d/%d complete]'
               % (status['done'], status['total']))


def find_archive_latency(channel, padding, frametype=None, state=None,
                         base=const.OMICRON_ARCHIVE):
    """Find the latency of Omicron file archival for the given channel

    Parameters
    ----------
    channel : `str`
        name of channel
    padding : `int`
        padding parameter for Omicron processing
    frametype : `str`, optional
        frame type ID for data frame files
    state : `str`, optional
        name of DQSegDB flag defining operational state for this channel
    base : `str`, optional
        base directory for Omicron archive

    Returns
    -------
    latency : `dict`
        a `dict` of `(ext, latency)` pairs for each file extension stored in
        the archive ('root', 'xml.gz')
    """
    ifo = channel[:2]
    obs = ifo[0]
    # find latest GPS time
    if state is None and frametype is None:
        raise ValueError("Please give one of `state` or `frametype`")
    if state is None:
        target = get_latest_data_gps(obs, frametype)
    else:
        target = get_latest_active_gps(state)
    target -= padding
    # find latest file
    latency = {}
    for ext in ['root', 'xml.gz']:
        f = find_latest_omicron_file(channel, base, ext=ext)
        end = file_segment(f)[1]
        latency[ext] = (int(target - end), f)
    return latency



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous utilities
"""

from __future__ import print_function

import re
import os
from subprocess import Popen, PIPE, CalledProcessError
from distutils.version import StrictVersion

from six import PY2

from . import const


def which(program):
    """Find full path of executable program
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file


def shell(cmd, stdout=PIPE, **kwargs):
    """Execute a commdand in a `subprocess`

    Returns
    -------
    stdout : `str`
        the output (`stdout`) of the command

    Raises
    ------
    subprocess.CalledProcessError
        if the command returns a non-zero exit code
    """
    if isinstance(cmd, (list, tuple)):
        cmdstr = ' '.join(cmd)
    else:
        cmdstr = cmd
    proc = Popen(cmd, stdout=stdout, **kwargs)
    out, err = proc.communicate()
    if proc.returncode:
        raise CalledProcessError(proc.returncode, cmdstr)
    return out


def get_output_directory(args):
    """Return the output directory as parsed from the command-line args
    """
    if args.gps is None and not args.output_dir:
        args.output_dir = os.path.join(const.OMICRON_PROD, args.group)
    elif not args.output_dir:
        start, end = args.gps
        args.output_dir = os.path.join(
            const.OMICRON_PROD, '%s-%d-%d' % (args.group, start, end))
    return os.path.abspath(args.output_dir)


# -- version comparison utilities

class OmicronVersion(StrictVersion):
    version_re = re.compile(r'^v(\d+)r(\d+) (p(\d+))? ([ab](\d+))?$',
                            re.VERBOSE)

    def parse(self, vstring):
        match = self.version_re.match(vstring)
        if not match:
            raise ValueError("invalid version number '%s'" % vstring)

        (major, minor, patch, prerelease, prerelease_num) = \
            match.group(1, 2, 4, 5, 6)

        if patch:
            self.version = tuple(map(int, [major, minor, patch]))
        else:
            self.version = tuple(map(int, [major, minor])) + (None,)

        if prerelease:
            self.prerelease = (prerelease[0], int(prerelease_num))
        else:
            self.prerelease = None

    def __str__(self):
        if self.version[2] is None:
            return 'v%sr%s' % (self.version[0], self.version[1])
        else:
            return 'v%sr%sp%s' % (self.version[0], self.version[1],
                                  self.version[2])

    if PY2:
        def __cmp__(self, other):
            if isinstance(other, str):
                other = OmicronVersion(other)
            return StrictVersion.__cmp__(self, other)
    else:
        def _cmp(self, other):
            if isinstance(other, str):
                other = OmicronVersion(other)
            return StrictVersion._cmp(self, other)


def get_omicron_version(executable=None):
    """Parse the version number from the Omicron executable path

    Parameters
    ----------
    executable : `str`
        path of Omicron executable

    Returns
    -------
    version : `str`
        the Omicron-format version string, e.g. `vXrY`

    Examples
    --------
    >>> get_omicron_version("/home/detchar/opt/virgosoft/Omicron/v2r1/Linux-x86_64/omicron.exe")
    'v2r1'
    """  # noqa: E501
    if executable:
        executable = os.path.abspath(executable)
        distdir = os.path.dirname(executable)
        versiondir = os.path.dirname(distdir)
        vstr = os.path.basename(versiondir)
    elif os.getenv('OMICRON_VERSION'):
        vstr = os.environ['OMICRON_VERSION']
    else:
        try:
            vstr = os.path.basename(os.environ['OMICRONROOT'])
        except KeyError as e:
            e.args = ('Cannot parse Omicron version from environment, '
                      'please specify the executable path',)
            raise
    return OmicronVersion(vstr)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for omicron.io
"""

from .. import (io, const)


def test_get_archive_filename():
    assert io.get_archive_filename('L1:GDS-CALIB_STRAIN', 0, 100) == (
        '%s/L1/GDS_CALIB_STRAIN_OMICRON/00000/'
        'L1-GDS_CALIB_STRAIN_OMICRON-0-100.xml.gz' % const.OMICRON_ARCHIVE)
    assert io.get_archive_filename(
            'L1:GDS-CALIB_STRAIN', 1234567890, 123, archive='/triggers',
            filetag='TEST-TAg', ext='root') == (
        '/triggers/L1/GDS_CALIB_STRAIN_TEST_TAg/12345/'
        'L1-GDS_CALIB_STRAIN_TEST_TAg-1234567890-123.root')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Setup the PyOmicron package
"""

import glob
import os.path
import sys

from setuptools import (setup, find_packages)

import versioneer
__version__ = versioneer.get_version()
cmdclass = versioneer.get_cmdclass()

# -- metadata ---------------

# set basic metadata
PACKAGENAME = 'pyomicron'
DISTNAME = 'pyomicron'
AUTHOR = 'Duncan Macleod'
AUTHOR_EMAIL = 'duncan.macleod@ligo.org'
LICENSE = 'GPLv3'

# read description
with open('README.md', 'rb') as f:
    longdesc = f.read().decode().strip()

# -- dependencies -----------

setup_requires = [
    'setuptools',
]
install_requires = [
    'dqsegdb2',
    'gwdatafind',
    'gwpy >= 0.14.1',
    'h5py',
    'htcondor',
    'ligo-segments',
    'lscsoft-glue >= 1.60.0',
    'MarkupPy',
    'numpy',
    'python-ligo-lw >= 1.4.0',
    'six',
]
tests_require = [
    'pytest',
]
if {'test'}.intersection(sys.argv):
    setup_requires.extend([
        'pytest_runner',
    ])

# -- content ----------------

# Use the find_packages tool to locate all packages and modules
packagenames = find_packages()

# glob for all scripts
scripts = glob.glob(os.path.join('bin', '*'))


# -- run setup --------------

setup(
    # metadata
    name=DISTNAME,
    provides=[PACKAGENAME],
    version=__version__,
    description="Python utilities and wrappers for GW Omicron",
    long_description=longdesc,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    cmdclass=cmdclass,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    # dependencies
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    # content
    packages=packagenames,
    scripts=scripts,
    include_package_data=True,
)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of PyOmicron.
#
# PyOmicron is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyOmicron is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyOmicron.  If not, see <http://www.gnu.org/licenses/>.

"""Data utilities for Omicron
"""

from __future__ import print_function

import os
import glob
import re
import shutil
import warnings
try:
    from urllib.parse import urlparse
except ImportError:  # python < 3
    from urlparse import urlparse

import gwdatafind
from gwdatafind.utils import (filename_metadata, file_segment)

from ligo.segments import (segment as Segment, segmentlist as SegmentList)

re_ll = re.compile(r'_(llhoft|lldetchar)\Z')
re_gwf_gps_epoch = re.compile(r'[-\/](?P<gpsepoch>\d+)$')

AGGREGATED_HOFT = {  # map aggregated h(t) type to short h(t) type
    'H1_HOFT_C00': 'H1_DMT_C00',
    'L1_HOFT_C00': 'L1_DMT_C00',
    'V1Online': 'V1_llhoft',
}


# -- utilities ----------------------------------------------------------------

def path_from_file_url(url):
    """Return the path of a `file://` URL
    """
    return urlparse(url).path


def _find_more_files(path):
    """Find more files similar to ``path`` by incrementing the GPS times

    This is mainly to find more files for a given (ifo, type) that aren't (yet)
    in the datafind server cache.

    Parameters
    ----------
    path : `str`
        the path of a file to use as the base, must conform to LIGO-T050017

    Returns
    -------
    morepaths : `list` of `str`
        the list of all files found by walking forward in time
    """
    # parse the GPS epoch from the path
    try:
        ngps = len(re_gwf_gps_epoch.search(
            os.path.dirname(path)).groupdict()['gpsepoch'])
    except AttributeError:
        # this failed, so we can't do anything
        return []
    else:
        found = []
        while True:
            s, e = file_segment(path)
            # replace start time with end time in filename
            new = path.replace('-{start}-'.format(start=s),
                               '-{end}-'.format(end=e))
            # and 5-digit GPS path in directory
            new = new.replace('{start5}/'.format(start5=str(s)[:ngps]),
                              '{end5}/'.format(end5=str(e)[:ngps]))
            # if this file doesn't exist, the previous file is what we want
            if not os.path.isfile(new):
                return found
            # otherwise keep going
            found.append(new)
            path = new


def ligo_low_latency_hoft_type(ifo, use_devshm=False):
    """Return the low-latency _h(t)_ frame type for the given interferometer

    Parameters
    ----------
    ifo : `str`
        prefix of IFO to use, e.g. 'L1'
    use_devshm : `bool`, optional
        use type in /dev/shm, default: `False`

    Returns
    -------
    frametype : `str`
        frametype to use for low-latency h(t)
    """
    if use_devshm:
        return '%s_llhoft' % ifo.upper()
    else:
        return '%s_DMT_C00' % ifo.upper()


def check_data_availability(obs, frametype, start, end):
    """Check for the full data availability for this frame type

    Parameters
    ----------
    obs : `str`
        the initial for the observatory
    frametype : `str`
        the name of the frame type for which to search
    start : `int`
        the GPS start time of this search
    end : `int`
        the GPS end time of this search

    Raises
    ------
    ValueError
        if gaps are found in the frame archive for the given frame type
    """
    return gwdatafind.find_urls(obs[0], frametype, start, end, on_gaps='error')


def write_cache(cache, outfile):
    if isinstance(outfile, str):
        with open(outfile, 'w') as fp:
            return write_cache(cache, fp).name

    for path in cache:
        obs, tag, seg = filename_metadata(path)
        print(
            '{0} {1} {2} {3} {4}'.format(obs, tag, seg[0], abs(seg), path),
            file=outfile)
    return outfile


# -- find files ---------------------------------------------------------------

def find_frames(obs, frametype, start, end, on_gaps='warn', **kwargs):
    """Find all frames for the given frametype in the GPS interval

    Parameters
    ----------
    obs : `str`
        the initial for the observatory
    frametype : `str`
        the name of the frame type for which to search
    start : `int`
        the GPS start time of this search
    end : `int`
        the GPS end time of this search
    **kwargs
        all other keyword arguments are passed directly to
        :func:`~gwdatafind.find_urls`

    Returns
    -------
    paths : `list` of `str`
        a list of GWF file pths
    """
    ll_kw = {key: kwargs.pop(key) for key in ('tmpdir', 'root',) if
             key in kwargs}

    cache = _find_frames_datafind(obs, frametype, start, end, on_gaps='ignore',
                                  **kwargs)

    # find more files for low-latency under /dev/shm (or similar)
    if re_ll.search(frametype):
        try:
            latest = file_segment(cache[-1])[1]
        except IndexError:
            latest = start
        if latest < end:
            cache.extend(find_ll_frames(obs, frametype, latest, end, **ll_kw))

    # handle missing files
    if on_gaps != 'ignore':
        seglist = SegmentList(map(file_segment, cache)).coalesce()
        missing = (SegmentList([Segment(start, end)]) - seglist).coalesce()
        msg = "Missing segments:\n%s" % '\n'.join(map(str, missing))
        if missing and on_gaps == 'warn':
            warnings.warn(msg)
        elif missing:
            raise RuntimeError(msg)

    return cache


def _find_frames_datafind(obs, frametype, start, end, **kwargs):
    kwargs.setdefault('urltype', 'file')
    cache = list(map(
        path_from_file_url,
        gwdatafind.find_urls(obs[0], frametype, start, end, **kwargs),
    ))

    # use latest frame to find more recent frames that aren't in
    # datafind yet, this is quite hacky, and isn't guaranteed to
    # work at any point, but it shouldn't break anything
    try:
        latest = cache[-1]
    except IndexError:  # no frames, `cache` is list()
        latestgps = start
    else:
        cache.extend(_find_more_files(latest))
        latestgps = file_segment(cache[-1])[1]

    # if we're searching for aggregated h(t), find more files
    # for the equivalent short h(t) type:
    if frametype in AGGREGATED_HOFT and latestgps < end:
        cache.extend(_find_frames_datafind(
            obs,
            AGGREGATED_HOFT[frametype],
            latestgps,
            end,
            **kwargs
        ))

    return cache


def find_ll_frames(ifo, frametype, start, end, root='/dev/shm', tmpdir=None):
    """Find all buffered low-latency frames in the given interval

    Parameters
    ----------
    ifo : `str`
        the IFO prefix, e.g. 'L1'
    frametype : `str`
        the frame type identifier, e.g. 'llhoft'
    start : `int`
        the GPS start time of this search
    end : `int`
        the GPS end time of this search
    root : `str`, optional
        the base root for the buffer, defaults to `/dev/shm`
    on_gaps : `str`, optional
        what to do when the found frames don't cover the full span, one of
        'warn', 'raise', or 'ignore'
    tmpdir : `str`, optional
        temporary directory into which to copy files from /dev/shm

        ..note::

           Caller is reponsible for deleting the direcotyr and its
           contents when done with it.

    Returns
    -------
    paths : `list` of `str`
        a list of GWF file pths

    .. warning::

       This method is not safe, given that the frames may disappear from
       the buffer before you have had a chance to read them

    """
    seg = Segment(start, end)
    cache = list(filter(lambda x: file_segment(x).intersects(seg),
                        _find_ll_frames(ifo, frametype, root=root)))
    if tmpdir:
        out = []
        if not os.path.isdir(tmpdir):
            os.makedirs(tmpdir)
        for path in cache:
            new = os.path.join(tmpdir, os.path.basename(path))
            shutil.copyfile(path, new)
            out.append(new)
        return out
    return cache


def _find_ll_frames(ifo, frametype, root='/dev/shm', ext='gwf'):
    obs = ifo[0]
    bits = frametype.rsplit('_', 1)
    basedir = os.path.join(root, *bits[::-1])
    globstr = os.path.join(basedir, '{obs}-{frametype}-*-*.{ext}'.format(
        obs=obs, frametype=frametype, ext=ext))
    # don't return the last file, as it might not have been fully written yet
    return sorted(glob.glob(globstr)[:-1])


# -- find latest file ---------------------------------------------------------

def get_latest_data_gps(obs, frametype):
    """Get the end GPS time of the latest available frame file

    Parameters
    ----------
    obs : `str`
        the initial for the observatory
    frametype : `str`
        the name of the frame type for which to search

    Returns
    -------
    gpstime : `int`
        the GPS time marking the end of the latest frame
    """
    try:
        latest = _find_latest_file(obs, frametype)
    except IndexError as e:
        e.args = ('No {0[0]}-{1} frames found'.format(obs, frametype),)
        raise
    # return end time of file as indicated by filename
    return int(file_segment(latest)[1])


def _find_latest_file(obs, frametype):
    # find latest low-latency file using glob
    if re_ll.search(frametype):
        return _find_ll_frames(obs, frametype)[-1]

    # find latest file for short h(t) type preferrably
    if frametype in AGGREGATED_HOFT:
        return _find_latest_file(obs, AGGREGATED_HOFT[frametype])

    # otherwise use gwdatafind to find the latest file it knows about
    latest = gwdatafind.find_latest(obs[0], frametype, urltype='file',
                                    on_missing='error')[-1]

    # and then find more
    try:
        return _find_more_files(latest)[-1]
    except IndexError:
        return latest
